from random import randint
from typing import ClassVar
from unittest import TestCase
from unittest.mock import patch


class Dragon:
    DAMAGE_MIN: ClassVar[int] = 5
    DAMAGE_MAX: ClassVar[int] = 20
    HEALTH_MIN: ClassVar[int] = 50
    HEALTH_MAX: ClassVar[int] = 100
    name: str
    health: int
    position_x: int
    position_y: int

    def __init__(self, name: str, /,
                 *, position_x: int = 0, position_y: int = 0,
                 ) -> None:
        self.name = name
        self.health = randint(self.HEALTH_MIN, self.HEALTH_MAX)
        self.set_position(x=position_x, y=position_y)

    def get_position(self) -> tuple[int,int]:
        return self.position_x, self.position_y

    def set_position(self, *, x: int, y: int) -> None:
        self.position_x = x
        self.position_y = y

    def change_position(self, *,
             left: int = 0, right: int = 0,
             up: int = 0, down: int = 0,
             ) -> None:
        current_x, current_y = self.get_position()
        new_x = current_x + right - left
        new_y = current_y + down - up
        self.set_position(x=new_x, y=new_y)

    def make_damage(self) -> int:
        return randint(self.DAMAGE_MIN, self.DAMAGE_MAX)

    def take_damage(self, damage: int, /) -> None:
        self.health -= damage

    def is_dead(self):
        if self.health <= 0:
            return True
        else:
            return False


class TestName(TestCase):
    def test_name_default(self):
        with self.assertRaises(TypeError):
            dragon = Dragon()  # noqa

    def test_name_positional(self):
        dragon = Dragon('MyName')
        self.assertEqual(dragon.name, 'MyName')

    def test_name_keyword(self):
        with self.assertRaises(TypeError):
            dragon = Dragon(name='MyName')  # noqa


class TestHealth(TestCase):
    def setUp(self):
        self.dragon = Dragon('MyName')

    def test_health_default(self):
        self.assertGreaterEqual(self.dragon.health, 50)
        self.assertLessEqual(self.dragon.health, 100)

    def test_health_default_many(self):
        for _ in range(10_000):
            self.dragon = Dragon('MyName')
            self.assertGreaterEqual(self.dragon.health, 50)
            self.assertLessEqual(self.dragon.health, 100)

    def test_health_default_patch(self):
        with patch('dragon.randint', return_value=74) as randint:  # noqa
            dragon = Dragon('MyName')
            self.assertEqual(dragon.health, 74)

    def test_health_hasattr_HEALTH_MIN(self):
        self.assertTrue(hasattr(Dragon, 'HEALTH_MIN'))
        self.assertEqual(Dragon.HEALTH_MIN, 50)

    def test_health_hasattr_HEALTH_MAX(self):
        self.assertTrue(hasattr(Dragon, 'HEALTH_MAX'))
        self.assertEqual(Dragon.HEALTH_MAX, 100)

    def test_health_isdead_when_health_positive(self):
        self.dragon.health = 1
        result = self.dragon.is_dead()
        self.assertFalse(result)

    def test_health_isdead_when_health_zero(self):
        self.dragon.health = 0
        result = self.dragon.is_dead()
        self.assertTrue(result)

    def test_health_isdead_when_health_negative(self):
        self.dragon.health = 0
        result = self.dragon.is_dead()
        self.assertTrue(result)

    def test_health_damage_take_default(self):
        with self.assertRaises(TypeError):
            self.dragon.take_damage()  # noqa

    def test_health_damage_take_positional(self):
        self.dragon.health = 3
        self.dragon.take_damage(2)
        self.assertEqual(self.dragon.health, 1)

    def test_health_damage_take_keyword(self):
        with self.assertRaises(TypeError):
            self.dragon.take_damage(damage=1)  # noqa


class TestPosition(TestCase):
    def setUp(self):
        self.dragon = Dragon('MyName')

    def test_position_init_default(self):
        self.assertEqual(self.dragon.position_x, 0)
        self.assertEqual(self.dragon.position_y, 0)

    def test_position_init_positional(self):
        with self.assertRaises(TypeError):
            dragon = Dragon('MyName', 1, 2)  # noqa

    def test_position_init_keyword(self):
        dragon = Dragon('MyName', position_x=1, position_y=2)
        self.assertEqual(dragon.position_x, 1)
        self.assertEqual(dragon.position_y, 2)

    def test_position_get(self):
        self.dragon.position_x = 1
        self.dragon.position_y = 2
        result = self.dragon.get_position()
        self.assertIsInstance(result, tuple)
        self.assertEqual(result[0], 1)
        self.assertEqual(result[1], 2)

    def test_position_set_default(self):
        with self.assertRaises(TypeError):
            self.dragon.set_position()  # noqa

    def test_position_set_positional(self):
        with self.assertRaises(TypeError):
            self.dragon.set_position(1)  # noqa
        with self.assertRaises(TypeError):
            self.dragon.set_position(1, 2)  # noqa

    def test_position_set_keyword(self):
        self.dragon.set_position(x=1, y=2)
        self.assertEqual(self.dragon.position_x, 1)
        self.assertEqual(self.dragon.position_y, 2)

    def test_position_change_positional(self):
        with self.assertRaises(TypeError):
            self.dragon.change_position(1)  # noqa
        with self.assertRaises(TypeError):
            self.dragon.change_position(1, 2)  # noqa
        with self.assertRaises(TypeError):
            self.dragon.change_position(1, 2, 3)  # noqa
        with self.assertRaises(TypeError):
            self.dragon.change_position(1, 2, 3, 4)  # noqa

    def test_position_change_keyword_left(self):
        self.dragon.position_x = 10
        self.dragon.position_y = 20
        self.dragon.change_position(left=1)
        self.assertEqual(self.dragon.position_x, 9)
        self.assertEqual(self.dragon.position_y, 20)

    def test_position_change_keyword_right(self):
        self.dragon.position_x = 10
        self.dragon.position_y = 20
        self.dragon.change_position(right=1)
        self.assertEqual(self.dragon.position_x, 11)
        self.assertEqual(self.dragon.position_y, 20)

    def test_position_change_keyword_up(self):
        self.dragon.position_x = 10
        self.dragon.position_y = 20
        self.dragon.change_position(up=1)
        self.assertEqual(self.dragon.position_x, 10)
        self.assertEqual(self.dragon.position_y, 19)

    def test_position_change_keyword_down(self):
        self.dragon.position_x = 10
        self.dragon.position_y = 20
        self.dragon.change_position(down=1)
        self.assertEqual(self.dragon.position_x, 10)
        self.assertEqual(self.dragon.position_y, 21)

    def test_position_change_keyword_horizontal(self):
        self.dragon.position_x = 10
        self.dragon.position_y = 20
        self.dragon.change_position(left=1, right=2)
        self.assertEqual(self.dragon.position_x, 11)
        self.assertEqual(self.dragon.position_y, 20)

    def test_position_change_keyword_vertical(self):
        self.dragon.position_x = 10
        self.dragon.position_y = 20
        self.dragon.change_position(up=1, down=2)
        self.assertEqual(self.dragon.position_x, 10)
        self.assertEqual(self.dragon.position_y, 21)

    def test_position_change_keyword_omnidirectional(self):
        self.dragon.position_x = 10
        self.dragon.position_y = 20
        self.dragon.change_position(left=1, right=2, up=3, down=4)
        self.assertEqual(self.dragon.position_x, 11)
        self.assertEqual(self.dragon.position_y, 21)


class TestDamage(TestCase):
    def setUp(self):
        self.dragon = Dragon('MyName')

    def test_damage_make(self):
        dmg = self.dragon.make_damage()
        self.assertIsInstance(dmg, int)
        self.assertGreaterEqual(dmg, 5)
        self.assertLessEqual(dmg, 20)

    def test_damage_hasattr_DAMAGE_MIN(self):
        self.assertTrue(hasattr(Dragon, 'DAMAGE_MIN'))
        self.assertEqual(Dragon.DAMAGE_MIN, 5)

    def test_damage_hasattr_DAMAGE_MAX(self):
        self.assertTrue(hasattr(Dragon, 'DAMAGE_MAX'))
        self.assertEqual(Dragon.DAMAGE_MAX, 20)
